unit WaveIOX;

{
Unit WaveIOX;
This unit reads WAV files of almost any format
and converts them into 32-bit floating point buffers.

The basic structure is from the WaveIO unit by by
Carlos Barbosa(delphi@carlosb.com) with various
extensions like the reader/converter from me.
You can get the original I/O routines from his
site (http://www.carlosb.com).

Have fun!

Tobybear
www.tobybear.de
tobybear@web.de

History:
1.0: initial release
1.1: fixed some conversion errors,
     thanks to Christian Knufinke for finding them!
1.2: many bugfixes here and there ...
1.3: more bugfixes, new high-level functions for single arrays
1.4: some of Thaddy's suggestions implemented: buffer size is now 16384,
     mmIO removed (unneeded), replaced CopyMemory with Move
     added better filesharing compatibility on re-open
     improved destructor with try..finally
1.5: tiny bugfix resulting in range error
1.6: TPCMWaveReader.SetPosition fixed, thanks to Vern Baker
}

interface

uses
  Windows, Classes, MMSystem, SysUtils, MSACMX, Dialogs;

type TSingleArray = array of single;

const
  MPEGLAYER3 = 85;
  buffer_size = 16384;

  //High-Level functions
  function LoadWAVFileMono(fn: string; var sr, ch, size: longint): pointer;
  function LoadWAVFile(fn: string; var sr, ch, size: longint): pointer;
  procedure SaveWAVFile(fn: string; fdata: pointer; sr, ch, bits, size: longint);
  procedure SaveWAVFileSeparateStereo(fn: string; fdata1, fdata2: pointer;
   sr, ch, bits, size: longint);
  procedure GetWAVFileInfo(fn: string; var sr, ch, size: longint);

  procedure LoadWAVFileMonoA(fn: string; var FloatData: TSingleArray; var sr, ch, size: longint);
  procedure LoadWAVFileA(fn: string; var FloatData: TSingleArray; var sr, ch, size: longint);
  procedure LoadWAVFileSeparateStereoA(fn: string; var FloatData_L, FloatData_R: TSingleArray;
   var sr, ch, size: longint);
  procedure SaveWAVFileA(fn: string; FloatData: TSingleArray; sr, ch, bits, size: longint);
  procedure SaveWAVFileSeparateStereoA(fn: string; FloatData_L, FloatData_R: TSingleArray;
   sr, ch, bits, size: longint);

type
  EWaveIOError = Exception;

  TWaveStream = class(TObject)
  private
    FFormat: PWaveFormatEx;
    FDataOffset, FSize: Longint;
    FullSize, FlSize:Longint;
    mm: hmmIO;
    pck, pckRIFF: TMMCKINFO;
    function GetPosition: longint;
    procedure SetPosition(Pos: longint);
    procedure CheckMMIOWave;
  public
    constructor Create(mmIO: hmmIO; WriteFormat: PWaveFormatEx);
    destructor Destroy; override;
    function Read(var Buffer; Count: longint): Longint; virtual;
    function Seek(Offset: longint; Origin: word): Longint; virtual;
    function Write(var Buffer; Count: longint): Longint; virtual;
    property Format: PWaveFormatEx read FFormat;
    property Position: Longint read GetPosition write SetPosition;
    property Size: Longint read FSize;
    property FloatSize:Longint read FlSize;
  end;

  TMemoryWaveStream = class(TWaveStream)
  public
   constructor Create(Memory: pointer; Size: longint;
    WriteFormat: PWaveFormatEx); virtual;
  end;

  TFileWaveStream = class(TWaveStream)
  public
   constructor Create(FileName: string; WriteFormat: PWaveFormatEx); virtual;
  end;

  TPCMWaveReader = class(TObject)
  private
    FPosition: longint;
    FStream: TWaveStream;
    FSize: longint;
    FBufferLength: longint;
    acmStream: integer;
    ash: TACMStreamHeader;
    dstwfx: PWaveFormatEx;
    PFullSize, PFlSize: longint;
    RawBuffer: pointer;
    RawBufferByteSize: longint;
    RawBufferSampleSize: longint;
    PCMBuffer: pointer;
    PCMBufferByteSize: longint;     // size of total buffer in bytes
    PCMBufferSampleSize: longint;   // size of converted samples in buffer
    PCMSamplesPerSample, PCMBufferSamplePos: longint;
    procedure AllocBuffers;
    procedure DestroyBuffers;
    procedure ReadSamples(Index: longint);
    procedure SetBufferLength(Value: longint);
    procedure SetPosition(Value: longint);
  public
    constructor Create(Stream: TWaveStream);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): longint;
    property Format: PWaveFormatEx read dstwfx;
    property BufferLength: longint read FBufferLength write SetBufferLength;
    property Position: longint read FPosition write SetPosition;
    property Size: longint read FSize;
    property Stream: TWaveStream read FStream;
    function BufferToFloat(makemono: boolean): pointer;
  end;

  TWavWriter = class(TObject)
  public
   Format: TWaveFormatEx;
   Stream: TFileWaveStream;
   constructor Create(fn: string; sr, ch, bits: longint);
   procedure WriteFloatData(p: pointer; size: longint);
   procedure WriteFloatDataSeparateStereo(p1, p2: pointer; size: longint);
   destructor Destroy; override;
  end;

implementation

function mmioFourCC(Chr1: Char; Chr2: Char; Chr3: Char; Chr4: Char): DWord;
begin
 Result := Integer(Chr1) + (Integer(Chr2) shl 8) + (Integer(Chr3) shl 16)
  + (Integer(Chr4) shl 24);
end;

constructor TWaveStream.Create(mmIO: hmmIO; WriteFormat: PWaveFormatEx);
begin
  inherited Create;
  filemode := 0;
  mm := mmIO;

  if (WriteFormat <> nil) then
  begin
    // Create the output file RIFF chunk of form type 'WAVE'.
    pckRIFF.fccType := mmioFOURCC('W','A','V','E');
    pckRIFF.cksize := 0;
    if (mmioCreateChunk(mm, @pckRIFF, MMIO_CREATERIFF) <> 0) then
      raise EWaveIOError.Create('Error 01: Cannot create chunk.');

    // We are now descended into the 'RIFF' chunk we just created.
    // * Now create the 'fmt ' chunk. Since we know the size of this chunk,
    // specify it in the MMCKINFO structure so MMIO doesn't have to seek
    // back and set the chunk size after ascending from the chunk.
    pck.ckid := mmioFOURCC('f','m','t',' ');
    pck.cksize := sizeof(WriteFormat^) + WriteFormat^.cbSize;
    if (mmioCreateChunk(mm, @pck, 0) <> 0) then
      raise EWaveIOError.Create('Error 02: Cannot create chunk.');
    // Write the variable length size.
    if (mmioWrite(mm, Pointer(WriteFormat), sizeof(WriteFormat^) +
     WriteFormat^.cbSize) <> sizeof(WriteFormat^) + WriteFormat^.cbSize) then
      raise EWaveIOError.Create('Error 03: Cannot write wave format.');
    GetMem(FFormat, sizeof(FFormat^));
    Move(WriteFormat^, FFormat^, Sizeof(FFormat^));

    // Ascend out of the 'fmt ' chunk,back into the 'RIFF' chunk.
    if (mmioAscend(mm, @pck, 0) <> 0) then
      raise EWaveIOError.Create('Error 04: Cannot ascend chunk.');

    // We are now descended into the 'RIFF' chunk we just created.
    // * Now create the 'data' chunk.
    pck.ckid := mmioFOURCC('d','a','t','a');
    pck.cksize := 0;
    if (mmioCreateChunk(mm, @pck, 0) <> 0) then
      raise EWaveIOError.Create('Error 05: Cannot create chunk.');
  end
  else
  begin
    CheckMMIOWave;
    // put in the beggining of the file
    mmioSeek(mm, FDataOffset, SEEK_SET);
  end;
end;

destructor TWaveStream.Destroy;
begin
  if (FFormat <> nil) then FreeMem(FFormat);
  mmioSeek(mm, 0, SEEK_END);
  mmioAscend(mm, @pck, 0);
  mmioAscend(mm, @pckRIFF, 0);
  mmioFlush(mm, 0);
  inherited Destroy;
end;

function TWaveStream.GetPosition: Longint;
begin
 Result := Seek(0, SEEK_CUR);
end;

procedure TWaveStream.SetPosition(Pos: Longint);
begin
 Seek(Pos, SEEK_SET);
end;

function TWaveStream.Read(var Buffer; Count: Longint): Longint;
var p: Pointer;
begin
 p := @Buffer;
 Result := mmioRead(mm, p, Count * FFormat^.nBlockAlign);
 if (Result = -1) then
  raise EWaveIOError.Create('Error 09: Cannot read from file.')
 else
  Result := Result div FFormat^.nBlockAlign;
end;

function TWaveStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
 if (Origin = SEEK_SET) then
  Result := (mmioSeek(mm, Offset * FFormat^.nBlockAlign +
   FDataOffset, Origin) - FDataOffset) div FFormat^.nBlockAlign
 else
  Result := (mmioSeek(mm, Offset * FFormat^.nBlockAlign,
   Origin) - FDataOffset) div FFormat^.nBlockAlign;
 if (Result < 0) then
  raise EWaveIOError.Create('Error 10: Cannot seek in file.');
end;

function TWaveStream.Write(var Buffer; Count: Longint): Longint;
begin
 Result := mmioWrite(mm, @Buffer,
  Count * FFormat^.nBlockAlign div FFormat^.nChannels);
 if (Result = -1) then
  raise EWaveIOError.Create('Error 11: Cannot write to file.')
 else
  Result := Result div FFormat^.nBlockAlign;
end;

procedure TWaveStream.CheckMMIOWave;
var mmIOInfo: TMMIOInfo;
    FormatTmp: TWaveFormatEx;
    ExtraAlloc: Word;
begin
 if (mmioDescend(mm, @pckRIFF, nil, 0) <> 0) then
  raise EWaveIOError.Create('Error 12: Invalid multimedia file!');

 if (pckRIFF.ckid <> FOURCC_RIFF)
  or (pckRIFF.fccType <> mmioFOURCC('W','A','V','E')) then
   raise EWaveIOError.Create('Error 13: Not a wave file!');

 // Search the input file for for the 'fmt ' chunk.     */
 pck.ckid := mmioFOURCC('f','m','t',' ');
 if (mmioDescend(mm, @pck, @pckRIFF, MMIO_FINDCHUNK) <> 0) then
  raise EWaveIOError.Create('Error 14: Cannot find ''fmt'' chunk!');

 // Expect the 'fmt' chunk to be at least as large as <PCMWAVEFORMAT>;
 // if there are extra parameters at the end,we'll ignore them */
 if (pck.cksize < 16) then
  raise EWaveIOError.Create('Error 15: Abnormal ''fmt'' size!');

 // Read the 'fmt ' chunk into <pcmWaveFormat>.*/
 if (mmioRead(mm, PChar(@FormatTmp), 16) <> 16) then
  raise EWaveIOError.Create('Error 16: Cannot read ''fmt'' chunk!');

 // Ok, allocate the waveformatex,but if its not pcm
 // format, read the next word,and thats how many extra
 // bytes to allocate.
 if (FormatTmp.wFormatTag = WAVE_FORMAT_PCM) or
    (FormatTmp.wFormatTag = 3) then
  ExtraAlloc:=0
 else
  // Read in length of extra bytes.
  if (mmioRead(mm, @ExtraAlloc, sizeof(ExtraAlloc)) <>
   sizeof(ExtraAlloc)) then
    raise EWaveIOError.Create('Error 17: Cannot read ''waveformatex'' length!');

 GetMem(FFormat, sizeof(FFormat^) + ExtraAlloc);
 Move(FormatTmp, FFormat^, SizeOf(FFormat^));
 FFormat^.cbSize := ExtraAlloc;

 if (ExtraAlloc <> 0) then
  if (mmioRead(mm, PChar(FFormat) + sizeof(FFormat^),
   ExtraAlloc) <> ExtraAlloc) then
    raise EWaveIOError.Create('Error 18: Cannot read ''waveformatex''!');

 if (FFormat^.wFormatTag = MPEGLAYER3) then
  raise EWaveIOError.Create('Error 19: MPEG Layer-3 compression is not supported.');

 // Ascend the input file out of the 'fmt ' chunk. */
 if (mmioAscend(mm, @pck, 0) <> 0) then
  raise EWaveIOError.Create('Error 20: Cannot ascend from ''fmt'' chunk!');

 // Do a nice little seek...
 if (mmioSeek(mm, pckRIFF.dwDataOffset + sizeof(FOURCC),
  SEEK_SET) = -1) then
  raise EWaveIOError.Create('Error 21: Cannot seek to data!');

 // Search the input file for the 'data' chunk.
 pck.ckid := mmioFOURCC('d','a','t','a');
 mmioDescend(mm, @pck, @pckRIFF, MMIO_FINDCHUNK);

 if (mmioGetInfo(mm, @mmioInfo, 0) <> 0) then
  raise EWaveIOError.Create('Error 22: Cannot get info!');
 FDataOffset := pck.dwDataOffset;
 FSize := pck.cksize div FFormat^.nBlockAlign;
 FlSize := FSize;
end;


//***************************************************************
//        TMemoryWaveStream
//***************************************************************

constructor TMemoryWaveStream.Create(Memory: Pointer; Size: Longint;
  WriteFormat: PWaveFormatEx);
var info: TMMIOINFO;
begin
 ZeroMemory(@info, sizeof(info));
 with info do
 begin
  pchBuffer := Memory;
  fccIOProc := FOURCC_MEM;
  cchBuffer := Size;
 end;
 // Initialization...
 mm := mmioOpen(nil, @info,MMIO_READ);
 if (mm = 0) then
  raise EWaveIOError.Create('Error 23: Cannot open memory stream.');
 inherited Create(mm, WriteFormat);
end;

//***************************************************************
//        TFileWaveStream
//***************************************************************

constructor TFileWaveStream.Create(FileName: String;
 WriteFormat: PWaveFormatEx);
var f2: TFileStream;
begin
 // Initialization...
 if (WriteFormat = nil) then
 begin
  f2 := TFileStream.Create(filename, fmOpenRead or fmShareDenyNone);
  fullsize := f2.size;
  f2.free;
  mm := mmioOpen(Pointer(FileName), nil, MMIO_READ or MMIO_ALLOCBUF);
 end else
  mm := mmioOpen(Pointer(FileName), nil, MMIO_CREATE or MMIO_READWRITE
   or MMIO_ALLOCBUF);   
 if (mm = 0) then
   raise EWaveIOError.Create('Error 25: Cannot open file stream.');
 inherited Create(mm, WriteFormat);
end;


//***************************************************************
//        TPCMWaveReader
//***************************************************************

constructor TPCMWaveReader.Create(Stream: TWaveStream);
begin
 inherited Create;
 FStream := Stream;
 GetMem(dstwfx, sizeof(dstwfx^));
//  FBufferLength:=4096;
 FBufferLength := 20000;
 if (FStream.Format^.wFormatTag <> WAVE_FORMAT_PCM) and
    (FStream.Format^.wFormatTag <> 3) then
 begin
  // prepare acm stream converter
  ZeroMemory(dstwfx,sizeof(dstwfx^));
  dstwfx^.wFormatTag := WAVE_FORMAT_PCM;
  if (acmFormatSuggest(0, FStream.FFormat, dstwfx, sizeof(dstwfx^),
   ACM_FORMATSUGGESTF_WFORMATTAG) <> 0) then
    raise EWaveIOError.Create('Error 26: Cannot suggest PCM format.');
  if (acmStreamOpen(PHACMSTREAM(@acmStream), 0, FStream.Format^, dstwfx^,
   nil, 0, 0, 0) <> 0) then
    raise EWaveIOError.Create('Error 27: Cannot open ACM stream.');
  AllocBuffers;
  // prepare buffers
  ZeroMemory(@ash, sizeof(ash));
  with ash do
  begin
   cbStruct := sizeof(ash);
   pbSrc := Pointer(RawBuffer);
   cbSrcLength := RawBufferByteSize;
   dwSrcUser := cbSrcLength;
   pbDst := Pointer(PCMBuffer);
   cbDstLength := PCMBufferByteSize;
   dwDstUser := cbDstLength;
  end;
  if (acmStreamPrepareHeader(acmStream, ash, 0) <> 0) then
   raise EWaveIOError.Create('Error 28: Cannot prepare headers.');
 end else dstwfx^ := FStream.Format^;
 if(FStream.Format^.wFormatTag <> WAVE_FORMAT_PCM) and
   (FStream.Format^.wFormatTag <> 3) then
   PCMSamplesperSample :=(PCMBufferByteSize * FStream.Format^.nBlockAlign)
     div(RawBufferByteSize * dstwfx^.nBlockAlign)
 else
   PCMSamplesPerSample := 1;
 FSize := PCMSamplesPerSample * FStream.Size;
 PCMBufferSampleSize := 0;
end;

destructor TPCMWaveReader.Destroy;
begin
 if (acmStream <> 0) then
 begin
  ash.cbSrcLength := RawBufferByteSize;
  acmStreamUnprepareHeader(acmStream, ash, 0);
  acmStreamClose(acmStream, 0);
 end;
 if (dstwfx <> nil) then FreeMem(dstwfx);
 DestroyBuffers;
 inherited Destroy;
end;

procedure TPCMWaveReader.AllocBuffers;
var ss, dd: DWord;
begin
 DestroyBuffers;
 // calc space needed for holding the decompression of one sample
 if (acmStreamSize(acmStream, FStream.Format^.nBlockAlign, dd,
  ACM_STREAMSIZEF_SOURCE) <> 0) then
   raise EWaveIOError.Create('Error 29: Cannot recommend an ACM stream size.');
 // calc minimum size block of the source
 if (acmStreamSize(acmStream, dd, ss, ACM_STREAMSIZEF_DESTINATION) <> 0) then
   raise EWaveIOError.Create('Error 30: Cannot recommend an ACM stream size.');
 // alloc source buffer(raw)
 RawBufferSampleSize := FBufferLength div dd;
 if (RawBufferSampleSize = 0) then RawBufferSampleSize := 1;
 RawBufferByteSize := RawBufferSampleSize * ss;
 GetMem(RawBuffer, RawBufferByteSize);
 // Alloc destination buffer(decompressed)
 if (acmStreamSize(acmStream, DWord(RawBufferByteSize),
   DWord(PCMBufferByteSize), ACM_STREAMSIZEF_SOURCE) <> 0) then
   raise EWaveIOError.Create('Error 31: Cannot recommend an ACM stream size.');
 GetMem(PCMBuffer, PCMBufferByteSize);
end;


procedure TPCMWaveReader.DestroyBuffers;
begin
 PCMBufferSampleSize := 0;
 if (PCMBuffer <> nil) then FreeMem(PCMBuffer);
 PCMBuffer := nil;
 if (RawBuffer <> nil) then FreeMem(RawBuffer);
 RawBuffer := nil;
end;

procedure TPCMWaveReader.SetBufferLength(Value: Longint);
begin
 if (Value <> FBufferLength) and(Value >= 1024) then
 begin
  FBufferLength := Value;
  if (PCMBuffer <> nil) then AllocBuffers;
 end;
end;

procedure TPCMWaveReader.SetPosition(Value: Longint);
begin
 if (Value <> FPosition) then
  if (Value >= 0) and (Value <= FSize) then
   FPosition := Value
  else
   raise EWaveIOError.Create('Error 32: Position out of bounds.');
end;

function TPCMWaveReader.Read(var Buffer; Count: longint): longint;
var pos, len: longint;
    posi, posf: longint;
begin
 if (Count < 1) then
 begin
  Result := 0;
  exit;
 end;
 if (FStream.Format^.wFormatTag = WAVE_FORMAT_PCM) or
    (FStream.Format^.wFormatTag = 3) then
 begin
  FStream.Position := FPosition;
  Result := FStream.Read(Buffer, Count);
  FPosition := FPosition + Result;
 end else
 begin
  // limit to wave size
  if (Count + FPosition >= FSize) then Count := FSize - FPosition;
  // use current buffer data if possible
  if (FPosition >= PCMBufferSamplePos) and
     (FPosition < PCMBufferSamplePos + PCMBufferSampleSize) then
  begin
   len := PCMBufferSamplePos + PCMBufferSampleSize - FPosition;
   if (len > Count) then len := Count;
   Move((PChar(PCMBuffer) + (FPosition - PCMBufferSamplePos) * dstwfx^.nBlockAlign)^,
    Buffer, len * dstwfx^.nBlockAlign);
   pos := len;
  end else pos:=0;
  // put next data
  while (pos < Count) do
  begin
   ReadSamples((FPosition + pos) div PCMSamplesPerSample);
   // calc range of current pcm buffer that can be used to fill request
   posi := FPosition + pos;
   if (FPosition + pos < PCMBufferSamplePos) then
    raise EWaveIOError.Create('Error 33: Position smaller than PCMBufferSamplePos');
   if (FPosition + Count > PCMBufferSamplePos + PCMBufferSampleSize) then
    posf := PCMBufferSamplePos + PCMBufferSampleSize
   else posf := FPosition + Count;
   len := posf - posi;
   // put pcm buffer data into target
   Move((PChar(PCMBuffer) + (posi - PCMBufferSamplePos) * dstwfx^.nBlockAlign)^,
   (PChar(@Buffer) + (posi - FPosition) * dstwfx^.nBlockAlign)^,
   len * dstwfx^.nBlockAlign);
   pos := pos + len;
  end;
  FPosition := FPosition + Count;
  Result := Count;
 end;
end;

procedure TPCMWaveReader.ReadSamples(Index: Longint);
begin
 FStream.Position := Index;
 ash.cbSrcLength := FStream.Read(RawBuffer^, RawBufferSampleSize) * FStream.Format^.nBlockAlign;
 if (acmStreamConvert(acmStream, ash, ACM_STREAMCONVERTF_BLOCKALIGN) <> 0) then
  raise EWaveIOError.Create('Error 34: Unable to convert sample.');
 PCMBufferSampleSize := ash.cbDstLengthUsed div dstwfx^.nBlockAlign;
 PCMBufferSamplePos := Index * PCMSamplesPerSample;
end;

function TPCMWaveReader.BufferToFloat(makemono: boolean): pointer;
var p: ^byte;
    p2: ^byte;
    ps, pf: ^single;
    fbuffer, buffer: pointer;
    l: longint;

procedure Convert(n: longint);
var
 cnt, cc, v: longint;
 v2: smallint;
 s: single;
 y, y2, y3: byte;
 by: byte;
begin
 cnt := 0;
 while(cnt <= format.nChannels * (n - 1)) do
 begin
 if (format.wBitsPerSample = 8) then
 begin
  for cc := 1 to format.nChannels do
  begin
   by := p2^; inc(p2);
   s := ((2 * by / (1 shl 8 - 1)) - 1);
   if (not makemono) then
   begin
    pf^ := s;
    inc(pf);
   end else
   begin
    pf^ := pf^ + s;
    if cc = format.nChannels then
    begin
     pf^ := pf^ / format.nChannels;
     inc(pf);
    end;
   end;
   inc(cnt);
  end;
 end else
 if (format.wBitsPerSample = 16) then
 begin
  for cc := 1 to format.nChannels do
  begin
   try
    y := p^; inc(p);
    y2 := p^; inc(p);
    v2 := y + y2 * (1 shl 8);
    s := v2;
    if s > 0 then s := s / (1 shl 15 - 1) else
     s := s / (1 shl 15);
   except
    s := 0;
   end;
   if (not makemono) then
   begin
    pf^ := s;
    inc(pf);
   end else
   begin
    pf^ := pf^ + s;
    if cc = format.nChannels then
    begin
     pf^ := pf^ / format.nChannels;
     inc(pf);
    end;
   end;
   inc(cnt);
  end;
 end else
 if (format.wBitsPerSample = 20) then
 begin
  for cc := 1 to format.nChannels do
  begin
   y := p^; inc(p);
   y2 := p^; inc(p);
   y3 := p^; inc(p);
   if not (format.nBlockAlign mod 3 = 0) then inc(p);
   v := (y + y2 * (1 shl 8) + y3 * (1 shl 16)) shr 4;
   if v >= 1 shl 19 then v := v - 1 shl 20;
   s := v;
   if s > 0 then s := s / (1 shl 19 - 1) else
    s := s / (1 shl 19);
   if (not makemono) then
   begin
    pf^ := s;
    inc(pf);
   end else
   begin
    pf^ := pf^ + s;
    if cc = format.nChannels then
    begin
     pf^ := pf^ / format.nChannels;
     inc(pf);
    end;
   end;
   inc(cnt);
  end;
 end else
 if (format.wBitsPerSample = 24) then
 begin
  for cc := 1 to format.nChannels do
  begin
   y := p^; inc(p);
   y2 := p^; inc(p);
   y3 := p^; inc(p);
   if not (format.nBlockAlign mod 3 = 0) then inc(p);
   v := y + y2 * (1 shl 8) + y3 * (1 shl 16);
   if v >= (1 shl 23) then v := v - (1 shl 24);
   s := v;
   if s > 0 then s := s / (1 shl 23 - 1) else
    s := s / (1 shl 23);
   if (not makemono) then
   begin
    pf^ := s;
    inc(pf);
   end else
   begin
    pf^ := pf^ + s;
    if cc = format.nChannels then
    begin
     pf^ := pf^ / format.nChannels;
     inc(pf);
    end;
   end;
   inc(cnt);
  end;
 end else
 if (format.wBitsPerSample = 32) then
 begin
  for cc := 1 to format.nChannels do
  begin
   s := ps^; inc(ps);
   if (not makemono) then
   begin
    pf^ := s;
    inc(pf);
   end else
   begin
    pf^ := pf^ + s;
    if cc = format.nChannels then
    begin
     pf^ := pf^ / format.nChannels;
     inc(pf);
    end;
   end;
   inc(cnt);
  end;
 end;
 end;
end;

begin
 if pfullsize > PFlSize then PFlSize := pfullsize;
 l := format.nChannels * PFlSize * sizeof(single);
 if makemono then l := l div format.nChannels;
 GetMem(FBuffer, l);
 GetMem(buffer, buffer_size * sizeof(smallint) * 4);
 FillChar(fbuffer^, l, 0);
 p := buffer;
 p2 := buffer;
 pf := fbuffer;
 l := Read(buffer^, BUFFER_SIZE);
 while (l > 0) do
 begin
  p := buffer;
  p2 := buffer;
  ps := buffer;
  convert(l);
  l := Read(buffer^, BUFFER_SIZE);
 end;
 if makemono then PFlSize := fsize
 else PFlSize := format.nChannels * fsize;
 FreeMem(buffer);
 result := fbuffer;
end;

{ TWavWriter }

constructor TWavWriter.Create(fn: string; sr, ch, bits: integer);
var p: PWaveFormatEx;
begin
 if FileExists(fn) then DeleteFile(fn);
 p := @format;
 FillChar(p^, sizeof(TWaveFormatEx), 0);
 Format.nChannels := ch;
 Format.cbSize := 0;
 Format.nSamplesPerSec := sr;
 if bits > 32 then bits := 32;
 Format.wBitsPerSample := bits;
 case bits of
 20, 24:begin Format.nBlockAlign := 3; Format.wFormatTag := WAVE_FORMAT_PCM end;
 32: begin Format.nBlockAlign := 4; Format.wFormatTag := 3 end;
 else begin
  if bits <= 8 then Format.nBlockAlign := 1
  else if bits <= 16 then Format.nBlockAlign := 2
  else if bits <= 24 then Format.nBlockAlign := 3
  else if bits < 32 then Format.nBlockAlign := 4;
  Format.wFormatTag := WAVE_FORMAT_PCM;
 end;
 end;
 Format.nBlockAlign := Format.nBlockAlign * ch;
 Format.nAvgBytesPerSec := sr * Format.nBlockAlign;
 stream := TFileWaveStream.Create(fn, p);
end;

destructor TWavWriter.Destroy;
begin
 try
  FreeAndNil(stream);
 except
 end; 
 sleep(100);
 inherited destroy;
end;

procedure TWavWriter.WriteFloatData(p: pointer; size: Integer);
var ps: psingle;
    l, li: longint;
    x: single;
begin
 ps := p;
 for l := 0 to size - 1 do
 begin
  if Format.wBitsPerSample = 32 then
  begin
   x := ps^;
   if x > 1 then x := 1 else if x < -1 then x := -1;
   li := longint((@x)^);
  end else
  begin
   x := ps^;
   if x > 1 then x := 1 else if x < -1 then x := -1;
   li := round(x * ((1 shl (Format.wBitsPerSample - 1)) - 1));
  end;
  stream.write(li, 1);
  inc(ps);
 end;
end;

procedure TWavWriter.WriteFloatDataSeparateStereo(p1,p2: pointer;
  size: Integer);
var ps, ps2: psingle;
    l, li: longint;
    x: single;
begin
 ps := p1;
 ps2 := p2;
 for l := 0 to size - 1 do
 begin
  if Format.wBitsPerSample = 32 then
  begin
   x := ps^;
   if x > 1 then x := 1 else if x < -1 then x := -1;
   li := longint((@x)^);
  end else
  begin
   x := ps^;
   if x > 1 then x := 1 else if x < -1 then x := -1;
   li := round(x * ((1 shl (Format.wBitsPerSample - 1)) - 1));
  end;
  stream.write(li, 1);
  inc(ps);
  if Format.wBitsPerSample = 32 then
  begin
   x := ps2^;
   if x > 1 then x := 1 else if x < -1 then x := -1;
   li := longint((@x)^);
  end else
  begin
   x := ps2^;
   if x > 1 then x := 1 else if x < -1 then x := -1;
   li := round(x * ((1 shl (Format.wBitsPerSample - 1)) - 1));
  end;
  stream.write(li, 1);
  inc(ps2);
 end;
end;

{ High-Level functions }

function LoadWAVFile(fn: string; var sr, ch, size: longint): pointer;
var wave: TFileWaveStream;
    pcmw: TPCMWaveReader;
begin
 wave := TFileWaveStream.Create(fn, nil);
 result := nil;
 try
  pcmw := TPCMWaveReader.Create(wave);
  sr := wave.format.nSamplesPerSec;
  ch := wave.format.nChannels;
  pcmw.pfullsize := wave.FullSize;
  result := pcmw.BufferToFloat(false);
  size := pcmw.pFlSize;
  pcmw.Free;
 finally
  wave.Free;
 end;
end;

procedure GetWAVFileInfo(fn:string; var sr, ch, size: longint);
var wave: TFileWaveStream;
begin
 wave := TFileWaveStream.Create(fn, nil);
 sr := wave.format.nSamplesPerSec;
 ch := wave.format.nChannels;
 size := wave.Size;
 wave.Free;
end;

procedure SaveWAVFile(fn: string; fdata: pointer; sr, ch, bits, size: longint);
var
  w: TFileWaveStream;
  ps: psingle;
  l, li: longint;
  t: TWaveFormatEx;
  p: PWaveFormatEx;
  x: single;
begin
 p := @t;
 FillChar(p^,sizeof(twaveformatex),0);
 t.nChannels := ch;
 t.nSamplesPerSec := sr;
 if bits > 32 then bits := 32;
 t.wBitsPerSample := bits;
 case bits of
 20, 24: begin t.nBlockAlign := 3; t.wFormatTag := WAVE_FORMAT_PCM end;
 32: begin t.nBlockAlign := 4; t.wFormatTag := 3 end;
 else begin
  if bits <= 8 then t.nBlockAlign := 1
  else if bits <= 16 then t.nBlockAlign := 2
  else if bits <= 24 then t.nBlockAlign := 3
  else if bits < 32 then t.nBlockAlign := 4;
  t.wFormatTag := WAVE_FORMAT_PCM;
 end;
 end;
 t.nBlockAlign := t.nBlockAlign * ch;
 t.nAvgBytesPerSec := sr * t.nBlockAlign;

 w := TFileWaveStream.Create(fn, p);
 ps := fdata;
 for l := 0 to size - 1 do
 begin
  if t.wBitsPerSample = 32 then
  begin
   x := ps^;
   if x > 1 then x := 1 else if x < -1 then x := -1;
   li := longint((@x)^);
  end else
  begin
   x := ps^;
   if x > 1 then x := 1 else if x < -1 then x := -1;
   li := round(x * ((1 shl (bits - 1)) - 1));
  end;
  w.write(li, 1);
  inc(ps);
 end;
 w.Free;
end;

procedure SaveWAVFileSeparateStereo(fn: string; fdata1, fdata2: pointer;
 sr, ch, bits, size: longint);
var
  w: TFileWaveStream;
  ps, ps2: psingle;
  l, li: longint;
  t: TWaveFormatEx;
  p: PWaveFormatEx;
  x: single;
begin
 p := @t;
 FillChar(p^, sizeof(TWaveFormatEx), 0);
 t.nChannels := ch;
 t.nSamplesPerSec := sr;
 if bits > 32 then bits := 32;
 t.wBitsPerSample := bits;
 case bits of
 20, 24: begin t.nBlockAlign := 3; t.wFormatTag := WAVE_FORMAT_PCM end;
 32: begin t.nBlockAlign := 4; t.wFormatTag := 3 end;
 else begin
  if bits <= 8 then t.nBlockAlign := 1
  else if bits <= 16 then t.nBlockAlign := 2
  else if bits <= 24 then t.nBlockAlign := 3
  else if bits < 32 then t.nBlockAlign := 4;
  t.wFormatTag := WAVE_FORMAT_PCM;
 end;
 end;
 t.nBlockAlign := t.nBlockAlign * ch;
 t.nAvgBytesPerSec := sr * t.nBlockAlign;

 w := TFileWaveStream.Create(fn, p);
 ps := fdata1;
 ps2 := fdata2;
 for l := 0 to size - 1 do
 begin
  if t.wBitsPerSample = 32 then
  begin
   x := ps^;
   if x > 1 then x := 1 else if x < -1 then x := -1;
   li := longint((@x)^);
  end else
  begin
   x := ps^;
   if x > 1 then x := 1 else if x < -1 then x := -1;
   li := round(x*((1 shl (bits - 1)) - 1));
  end;
  w.write(li, 1);
  inc(ps);
  if t.wBitsPerSample = 32 then
  begin
   x := ps2^;
   if x > 1 then x := 1 else if x < -1 then x := -1;
   li := longint((@x)^);
  end else
  begin
   x := ps2^;
   if x > 1 then x := 1 else if x < -1 then x := -1;
   li := round(x*((1 shl (bits - 1)) - 1));
  end;
  w.write(li, 1);
  inc(ps2);
 end;
 w.Free;
end;

function LoadWAVFileMono(fn: string; var sr, ch, size: longint): pointer;
var wave: TFileWaveStream;
    pcmw: TPCMWaveReader;
begin
 wave := TFileWaveStream.Create(fn, nil);
 try
  pcmw := TPCMWaveReader.Create(wave);
  sr := wave.format.nSamplesPerSec;
  pcmw.pfullsize := wave.FullSize;
  ch := 1;
  result := pcmw.BufferToFloat(true);
  size := pcmw.pFlSize;
  pcmw.Free;
 finally
  wave.Free;
 end; 
end;

procedure LoadWAVFileMonoA(fn: string; var FloatData: TSingleArray; var sr, ch, size: longint);
var ps, p: psingle;
    i: integer;
begin
 p := LoadWAVFileMono(fn, sr, ch, size);
 Setlength(FloatData, size);
 ps := p;
 for i := 0 to size - 1 do
 begin
  FloatData[i] := ps^;
  inc(ps);
 end;
 FreeMem(p);
end;

procedure LoadWAVFileA(fn: string; var FloatData: TSingleArray; var sr, ch, size: longint);
var ps, p: psingle;
    i: integer;
begin
 p := LoadWAVFile(fn, sr, ch, size);
 Setlength(FloatData, size);
 ps := p;
 for i := 0 to size - 1 do
 begin
  FloatData[i] := ps^;
  inc(ps);
 end;
 FreeMem(p);
end;

procedure LoadWAVFileSeparateStereoA(fn: string; var FloatData_L, FloatData_R: TSingleArray;
 var sr, ch, size: longint);
var ps, p: psingle;
    i: integer;
begin
 p := LoadWAVFile(fn, sr, ch, size);
 size := size div 2;
 SetLength(FloatData_L, size);
 SetLength(FloatData_R, size);
 ps := p;
 for i := 0 to (size div 2) - 1 do
 begin
  FloatData_L[i] := ps^;
  inc(ps);
  if ch = 2 then
  begin
   FloatData_R[i] := ps^;
   inc(ps);
  end else
  begin
   FloatData_R[i] := ps^;
  end;
 end;
 FreeMem(p);
end;

procedure SaveWAVFileA(fn: string; FloatData: TSingleArray; sr, ch, bits, size: longint);
begin
 SaveWAVFile(fn, psingle(FloatData), sr, ch, bits, size);
end;

procedure SaveWAVFileSeparateStereoA(fn: string; FloatData_L, FloatData_R: TSingleArray;
 sr, ch, bits, size: longint);
begin
 SaveWAVFileSeparateStereo(fn, psingle(FloatData_L), psingle(FloatData_R), sr, ch, bits, size);
end;

{
WAVEFORMATEX

The WAVEFORMATEX structure defines the format of
waveform-audio data. Only format information common
to all waveform-audio data formats is included in this
structure. For formats that require additional information,
this structure is included as the first member in another
structure, along with the additional information.

Formats that support more than two channels or sample sizes
of more than 16 bits can be described in a
WAVEFORMATEXTENSIBLE structure, which includes the
WAVEFORMAT structure.

    WORD  wFormatTag;
    WORD  nChannels;
    DWORD nSamplesPerSec;
    DWORD nAvgBytesPerSec;
    WORD  nBlockAlign;
    WORD  wBitsPerSample;
    WORD  cbSize;

Members

wFormatTag
Waveform-audio format type. Format tags are registered with
Microsoft Corporation for many compression algorithms.
A complete list of format tags can be found in the Mmreg.h
header file. For one- or two-channel PCM data, this value
should be WAVE_FORMAT_PCM. When this structure is included
in a WAVEFORMATEXTENSIBLE structure, this value must be
WAVE_FORMAT_EXTENSIBLE.

nChannels
Number of channels in the waveform-audio data. Monaural
data uses one channel and stereo data uses two channels.

nSamplesPerSec
Sample rate, in samples per second (hertz). If wFormatTag
is WAVE_FORMAT_PCM, then common values for nSamplesPerSec
are 8.0 kHz, 11.025 kHz, 22.05 kHz, and 44.1 kHz. For
non-PCM formats, this member must be computed according
to the manufacturer's specification of the format tag.

nAvgBytesPerSec
Required average data-transfer rate, in bytes per second,
for the format tag. If wFormatTag is WAVE_FORMAT_PCM,
nAvgBytesPerSec should be equal to the product of
nSamplesPerSec and nBlockAlign. For non-PCM formats,
this member must be computed according to the
manufacturer's specification of the format tag.

nBlockAlign
Block alignment, in bytes. The block alignment is the
minimum atomic unit of data for the wFormatTag format
type. If wFormatTag is WAVE_FORMAT_PCM or
WAVE_FORMAT_EXTENSIBLE, nBlockAlign must be equal
to the product of nChannels and wBitsPerSample divided
by 8 (bits per byte). For non-PCM formats, this
member must be computed according to the manufacturer's
specification of the format tag.
Software must process a multiple of nBlockAlign bytes of
data at a time. Data written to and read from a device
must always start at the beginning of a block. For
example, it is illegal to start playback of PCM data
in the middle of a sample (that is, on a non-block-aligned
boundary).

wBitsPerSample
Bits per sample for the wFormatTag format type. If
wFormatTag is WAVE_FORMAT_PCM, then wBitsPerSample
should be equal to 8 or 16. For non-PCM formats, this
member must be set according to the manufacturer's
specification of the format tag. If wFormatTag is
WAVE_FORMAT_EXTENSIBLE, this value can be any integer
multiple of 8 and represents the container size, not
necessarily the sample size; for example, a 20-bit sample
size is in a 24-bit container. Some compression schemes
cannot define a value for wBitsPerSample, so this member
can be 0.

cbSize
Size, in bytes, of extra format information appended to
the end of the WAVEFORMATEX structure. This information
can be used by non-PCM formats to store extra attributes
for the wFormatTag. If no extra information is required
by the wFormatTag, this member must be set to 0.
For WAVE_FORMAT_PCM formats (and only WAVE_FORMAT_PCM
formats), this member is ignored. When this structure
is included in a WAVEFORMATEXTENSIBLE structure, this
value must be at least 22.
}

end.


